/* Glyph class
 *
 * CMSC498B
 * Project 4
 * April 25, 2002
 *
 * Chad Ellinger
 * Developed with JDK 1.3.1_02 in Windows XP
 */

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;

/**
 * A <code>Glyph</code> object is a base class for all moving objects on the game screen.
 * Glyphs have dimension, direction, rotational speed, linear speed and acceleration.
 * Glyphs are also aware of the game's screen width and height. If the current position of
 * the Glyph is clipped by one of the four borders of the screen, the Glyph is also drawn on
 * the opposite side of the screen.
 */
public class Glyph{
    protected double x = 0.0, y = 0.0;
    protected int width = 0, height = 0;
    protected int screenWidth = 0, screenHeight = 0;
    protected int distanceTraveled = 0;
    protected double rotateSpeed = 0.0, linearSpeed = 30.0, maxSpeed = 100.0, accelRate = -1;
    protected double xLinearSpeed = -1, yLinearSpeed = -1;
    protected long time = 0, oldTime = 0;
    protected double deltaTime = 0;
    protected boolean moving = true, accel = false, selected = false;
    protected double direction = 0, angle = 0;
    protected Color glyphColor = Color.white;
    protected boolean topClip = false, rightClip = false, bottomClip = false, leftClip = false;
    protected AffineTransform at = new AffineTransform();
    protected Shape glyphShape, transformedShape;
    protected Area glyphArea = new Area();

    /**
     * Constructs a new Glyph.
     */
    public Glyph(){
    }

    /**
     * Returns true if given area collides with Glyph's area.
     *
     * @param testArea Area to test with current Glyph's area.
     */
    public boolean collision(Area testArea){
	Area tempArea = (Area)glyphArea.clone();
	tempArea.intersect(testArea);
	return !tempArea.isEmpty();
    }

    /**
     * Advances the glyph one step and draws it at its new position and rotation.
     */
    public void draw(Graphics2D g2){
	glyphArea.reset();
	at.setToIdentity();
	if (moving){
	    advanceGlyph();
	}
	at.translate((int)this.x, (int)this.y);
	at.rotate(angle, (double)this.width / 2, (double)this.height / 2);
       	g2.setTransform(at);
	transformedShape = at.createTransformedShape(glyphShape);
	glyphArea.add(new Area(transformedShape));
	drawGlyph(g2);

	leftClip = transformedShape.getBounds2D().getMinX() < 0;
	rightClip = transformedShape.getBounds2D().getMaxX() > screenWidth;
	topClip = transformedShape.getBounds2D().getMinY() < 0;
	bottomClip = transformedShape.getBounds2D().getMaxY() > screenHeight;

	if (leftClip){
	    at.setToIdentity();
	    at.translate((int)x + screenWidth, (int)y);
	    at.rotate(angle, (double)(this.width / 2), (double)(this.height / 2));
	    g2.setTransform(at);
	    glyphArea.add(new Area(at.createTransformedShape(glyphShape)));
	    drawGlyph(g2);
	}
	if (rightClip){
	    at.setToIdentity();
	    at.translate((int)x - screenWidth, (int)y);
	    at.rotate(angle, (double)(this.width / 2), (double)(this.height / 2));
	    g2.setTransform(at);
	    glyphArea.add(new Area(at.createTransformedShape(glyphShape)));
	    drawGlyph(g2);
	}
	if (topClip){
	    at.setToIdentity();
	    at.translate((int)x, (int)y + screenHeight);
	    at.rotate(angle, (double)(this.width / 2), (double)(this.height / 2));
	    g2.setTransform(at);
	    glyphArea.add(new Area(at.createTransformedShape(glyphShape)));
	    drawGlyph(g2);
	}
	if (bottomClip){
	    at.setToIdentity();
	    at.translate((int)x, (int)y - screenHeight);
	    at.rotate(angle, (double)(this.width / 2), (double)(this.height / 2));
	    g2.setTransform(at);
	    glyphArea.add(new Area(at.createTransformedShape(glyphShape)));
	    drawGlyph(g2);
	}
	if (leftClip && topClip){
	    at.setToIdentity();
	    at.translate((int)x + screenWidth, (int)y + screenHeight);
	    at.rotate(angle, (double)(this.width / 2), (double)(this.height / 2));
	    g2.setTransform(at);
	    glyphArea.add(new Area(at.createTransformedShape(glyphShape)));
	    drawGlyph(g2);
	}
	if (rightClip && topClip){
	    at.setToIdentity();
	    at.translate((int)x - screenHeight, (int)y + screenHeight);
	    at.rotate(angle, (double)(this.width / 2), (double)(this.height / 2));
	    g2.setTransform(at);
	    glyphArea.add(new Area(at.createTransformedShape(glyphShape)));
	    drawGlyph(g2);
	}
	if (leftClip && bottomClip){
	    at.setToIdentity();
	    at.translate((int)x + screenWidth, (int)y - screenHeight);
	    at.rotate(angle, (double)(this.width / 2), (double)(this.height / 2));
	    g2.setTransform(at);
	    glyphArea.add(new Area(at.createTransformedShape(glyphShape)));
	    drawGlyph(g2);
	}
	if (rightClip && bottomClip){
	    at.setToIdentity();
	    at.translate((int)x - screenWidth, (int)y - screenHeight);
	    at.rotate(angle, (double)(this.width / 2), (double)(this.height / 2));
	    g2.setTransform(at);
	    glyphArea.add(new Area(at.createTransformedShape(glyphShape)));
	    drawGlyph(g2);
	}
	return;
    }

    /**
     * Returns the current angle of the Glyph.
     */
    public double getAngle(){
	return angle;
    }

    /**
     * Returns the current Area of the Glyph.
     */
    public Area getArea(){
	return glyphArea;
    }

    /**
     * Returns the Glyph's color.
     */
    public Color getColor(){
	return glyphColor;
    }

    /**
     * Returns the center point of the Glyph.
     */
    public Point2D.Double getCenterPoint(){
	return new Point2D.Double(this.x + this.width / 2, this.y + this.height / 2);
    }

    /**
     * Returns the radial direction in which the Glyph is moving.
     */
    public double getDirection(){
	return direction;
    }

    /**
     * Returns the total distance the Glyph has traveled since creation.
     */
    public int getDistanceTraveled(){
	return distanceTraveled;
    }

    /**
     * Returns the height of the Glyph.
     */
    public int getHeight(){
	return height;
    }

    /**
     * Returns the Glyph's current linear speed.
     */
    public double getLinearSpeed(){
	return linearSpeed;
    }

    /**
     * Returns the Glyph's current Rotational speed.
     */
    public double getRotateSpeed(){
	return rotateSpeed;
    }

    /**
     * Returns the screen height that the Glyph is currently aware of.
     */
    public int getScreenHeight(){
	return screenHeight;
    }

    /**
     * Returns the screen width that the Glyph is currently aware of.
     */
    public int getScreenWidth(){
	return screenWidth;
    }

    /**
     * Returns true if Glyph is selected, false otherwise.
     */
    public boolean getSelected(){
	return selected;
    }

    /**
     * Returns the width of the Glyph.
     */
    public int getWidth(){
	return width;
    }

    /**
     * Returns the X-coordinate of the Glyph.
     */
    public double getX(){
	return x;
    }

    /**
     * Returns the Y-coordinate of the Glyph.
     */
    public double getY(){
	return y;
    }

    /**
     * Returns true if Glyph is accelerating, false otherwise.
     */
    public boolean isAccel(){
	return accel;
    }

    /**
     * Returns true if Glyph is moving, false otherwise.
     */
    public boolean isMoving(){
	return moving;
    }

    /**
     * Sets the Glyph to accelerate or not.
     */
    public void setAccel(boolean newAccel){
	accel = newAccel;
	return;
    }

    /**
     * Sets the angle to which the Glyph is rotated.
     */
    public void setAngle(double angle){
	this.angle = angle;
	return;
    }

    /**
     * Sets the Glyph's color.
     */
    public void setColor(Color newColor){
	glyphColor = newColor;
	return;
    }

    /**
     * Sets the Glyph's direction.
     */
    public void setDirection(double direction){
	this.direction = direction;
	return;
    }

    /**
     * Sets the Glyph's height.
     */
    public void setHeight(int height){
	this.height = height;
	return;
    }

    /**
     * Sets the Glyph's linear speed.
     */
    public void setLinearSpeed(double linearSpeed){
	this.linearSpeed = linearSpeed;
	return;
    }

    /**
     * Sets the Glyph to move or not.
     */
    public void setMoving(boolean moving){
	if (moving){
	    time = 0;
	}
	this.moving = moving;
	return;
    }

    /**
     * Sets the Glyph's rotational speed.
     */
    public void setRotateSpeed(double rotateSpeed){
	this.rotateSpeed = rotateSpeed;
	return;
    }

    /**
     * Sets the screen height to test clipping.
     */
    public void setScreenHeight(int height){
	this.screenHeight = height;
	return;
    }

    /**
     * Sets the screen width to test clipping.
     */
    public void setScreenWidth(int width){
	this.screenWidth = width;
	return;
    }

    /**
     * Sets the Glyph to be selected or not.
     */
    public void setSelected(boolean selected){
	this.selected = selected;
	return;
    }

    /**
     * Sets the Glyph's width.
     */
    public void setWidth(int width){
	this.width = width;
	return;
    }

    /**
     * Sets the Glyph's X-coordinate.
     */
    public void setX(double x){
	this.x = x;
	return;
    }

    /**
     * Sets the Glyph's Y-coordinate.
     */
    public void setY(double y){
	this.y = y;
	return;
    }

    protected void advanceGlyph(){
	if (time == 0){
	    oldTime = time = System.currentTimeMillis();
	    deltaTime = 0;
	}
	else{
	    oldTime = time;
	    time = System.currentTimeMillis();
	    deltaTime = (double)(time - oldTime)/1000;
	}
	if (accel){
	    linearSpeed = Math.min(maxSpeed, linearSpeed + accelRate*deltaTime);
	}
	else{
	    linearSpeed = Math.max(0.0, linearSpeed - accelRate*deltaTime);
	}
	xLinearSpeed = linearSpeed*Math.cos(direction);
	yLinearSpeed = linearSpeed*Math.sin(direction);
	x += xLinearSpeed*deltaTime;
	y += yLinearSpeed*deltaTime;
	distanceTraveled += (int)(linearSpeed*deltaTime);
	this.setAngle(angle + rotateSpeed*deltaTime);
	if (x > screenWidth){
	    x -= screenWidth;
	}
	else if ((x + width) < 0){
	    x += screenWidth;
	}
	if (y > screenHeight){
	    y -= screenHeight;
	}
	else if ((y + height) < 0){
	    y += screenHeight;
	}
	return;
    }

    protected void drawGlyph(Graphics2D g2){
	g2.setColor(glyphColor);
	g2.fill(glyphShape);
	if (selected){
	    g2.setColor(Color.cyan);
	    g2.draw(glyphShape);
	}
    }
}
